import datetime as dt
from pathlib import Path

import pandas as pd

from data_downloader import parse_urls


class SentinelOrbit:
    """a class to retrieve Sentinel-1 orbit data links.

    Examples
    --------
    >>> from data_downloader import downloader, services
    >>> from pathlib import Path

    >>> folder_cal = Path("/media/data/aux_cal")  # specify the folder to save aux_cal
    >>> folder_preorb = Path("/media/data/poeorb")  # specify the folder to save aux_poeorb

    init SentinelOrbit:

    >>> s1_orbit = services.SentinelOrbit()

    Get all aux_cal data links of S1A and S1B and download them:

    >>> urls_cal = s1_orbit.cal_urls(platform='all')
    >>> downloader.async_download_datas(urls_cal, folder=folder_cal)

    Get all precise orbit data links of S1A during 20210101-20220301 and download them:

    >>> urls_preorb = s1_orbit.poeorb_urls(
    >>>    date_start="20210101", date_end="20220301", platform="S1A"
    >>> )
    >>> downloader.download_datas(urls_preorb, folder=folder_preorb)
    """

    def __init__(
        self,
        home_aux_cal: str = "https://s1qc.asf.alaska.edu/aux_cal/",
        home_preorb: str = "https://s1qc.asf.alaska.edu/aux_poeorb/",
    ) -> None:
        """init SentinelOrbit.

        Parameters
        ----------
        home_aux_cal : str
            home url of aux_cal, default is "https://s1qc.asf.alaska.edu/aux_cal/".
        home_preorb : str
            home url of aux_poeorb, default is "https://s1qc.asf.alaska.edu/aux_poeorb/".
        """
        self.home_aux_cal = home_aux_cal
        self.home_preorb = home_preorb

    def cal_urls(self, platform="all"):
        """filter files from urls of aux_cal by platform.

        Parameters
        ----------
        platform : str, one of ['S1A', 'S1B','all']
            platform of satellite. should be one of ['S1A', 'S1B','all']
        """
        urls = parse_urls.from_html(self.home_aux_cal)
        if platform in ["S1A", "S1B", "all"]:
            if platform == "all":
                platform = ["S1A", "S1B"]
            else:
                platform = [platform]
        else:
            raise ValueError("platform must be one of ['S1A', 'S1B','all']")

        _urls = [i for i in urls if Path(i).suffix == ".SAFE"]
        urls_filter = [i for i in _urls if Path(i).stem[:3] in platform]

        return urls_filter

    def poeorb_urls(
        self,
        date_start: str,
        date_end: str,
        platform: str = "all",
    ):
        """filter files from urls of aux_poeorb (precise orbit) by date and platform.

        Parameters
        ----------
        date_start, date_end : str
            start/end date to filter, can be any format that can be converted by
            pd.to_datetime (e.g. '20210101', '2021-01-01', '2021/01/01').
        platform : str, one of ['S1A', 'S1B','all']
            platform of satellite. should be one of ['S1A', 'S1B','all']
        """
        if platform in ["S1A", "S1B", "all"]:
            if platform == "all":
                platform = ["S1A", "S1B"]
            else:
                platform = [platform]
        else:
            raise ValueError("platform must be one of ['S1A', 'S1B','all']")

        date_start = pd.to_datetime(date_start).date()
        date_end = pd.to_datetime(date_end).date()

        urls = parse_urls.from_html(self.home_preorb)
        _urls = [i for i in urls if Path(i).suffix == ".EOF"]
        urls_filter = []
        for i in _urls:
            name = Path(i).stem
            dt_i = pd.to_datetime(name.split("_")[-1]).date() - dt.timedelta(days=1)

            if name[:3] in platform and date_start <= dt_i <= date_end:
                urls_filter.append(i)

        return urls_filter



import re
from datetime import datetime
from pathlib import Path
import numpy as np
import pandas as pd

from data_downloader import parse_urls


class LiCSARService:
    """a class to retrieve LiCSAR data 

    Example:
    --------
    >>> from pathlib import Path
    >>> import pandas as pd
    >>> from data_downloader import downloader, services
    >>> # specify the folder to save data
    >>> home_dir = Path("/Volumes/Data/LiCSAR/106D_05248_131313/")
    >>> pair_dir = home_dir / "GEOC"

    init LiCSARService by frame id and download all metadata files

    >>> licsar = services.LiCSARService("106D_05248_131313")
    >>> downloader.download_datas(licsar.meta_urls, folder=home_dir, desc="Metadata")

    generate mask data by primary_dates, secondary_dates and day span

    >>> mask = (licsar.primary_dates>pd.to_datetime("2019-01-01")) & (licsar.primary_dates<pd.to_datetime("2019-12-31")) & (licsar.days < 12 * 5 + 1)

    download interferograms and coherence files filtered by mask

    >>> downloader.download_datas(licsar.ifg_urls[mask].values, folder=pair_dir, desc="Interferogram")
    >>> downloader.download_datas(licsar.coh_urls[mask], folder=pair_dir, desc="Coherence")
    """

    def __init__(
        self,
        frame_id: str,
        root_url: str = "https://gws-access.jasmin.ac.uk/public/nceo_geohazards/LiCSAR_products",
    ):
        """init LiCSARService.

        Parameters:
        -----------
        frame_id : str
            frame id of LiCSAR.
        root_url : str
            root url of LiCSAR. default is "https://gws-access.jasmin.ac.uk/public/nceo_geohazards/LiCSAR_products".
        """
        self.frame_id = frame_id
        self.track_id = self._parse_track_id()
        self.home_url = f"{root_url}/{self.track_id}/{self.frame_id}"

        self._pairs, self._ifg_urls, self._coh_urls = self._retrieve_pairs_urls()

    def __repr__(self):
        return f"LiCSAR(frame_id={self.frame_id}, count={len(self.pairs)})"

    def __str__(self) -> str:
        return f"LiCSAR(frame_id={self.frame_id}, count={len(self.pairs)})"

    def __len__(self):
        return len(self.pairs)

    def _parse_track_id(self):
        """parse track id from frame id."""
        track_id = str(int(self.frame_id[0:3]))
        return track_id

    def _retrieve_pairs_urls(self):
        """retrieve pairs of LiCSAR."""
        url = f"{self.home_url}/interferograms/"
        page_urls = parse_urls.from_html(url, url_depth=1)
        pairs = []
        ifg_urls = []
        coh_urls = []
        for i in page_urls:
            re_result = re.findall(r"\d{8}_\d{8}", i)
            if re_result:
                pairs.append(re_result[0])
                ifg_urls.append(f"{i}/{re_result[0]}.geo.unw.tif")
                coh_urls.append(f"{i}/{re_result[0]}.geo.cc.tif")
        return pairs, ifg_urls, coh_urls

    @property
    def pairs(self)->np.ndarray:
        """all available pairs of given frame id."""
        return np.array(self._pairs)

    @property
    def ifg_urls(self)->pd.Series:
        """interferogram urls of given frame id."""
        df = pd.Series(self._ifg_urls, name="ifg_urls", index=self._pairs)
        return df

    @property
    def coh_urls(self)->pd.Series:
        """coherence urls of given frame id."""
        df = pd.Series(self._coh_urls, name="coh_urls", index=self._pairs)
        return df

    @property
    def urls(self)->pd.DataFrame:
        """all urls, including interferogram and coherence, of given frame id."""
        urls = pd.concat([self.ifg_urls, self.coh_urls], axis=1)
        return urls

    @property
    def primary_dates(self)->np.ndarray:
        """primary dates of pairs for given frame id."""
        primary_dates = [
            datetime.strptime(i.split("_")[0], "%Y%m%d") for i in self.pairs
        ]
        return np.array(primary_dates, dtype="datetime64[D]")

    @property
    def secondary_dates(self)->np.ndarray:
        """secondary dates of pairs for given frame id."""
        secondary_dates = [
            datetime.strptime(i.split("_")[1], "%Y%m%d") for i in self.pairs
        ]
        return np.array(secondary_dates, dtype="datetime64[D]")

    @property
    def days(self)->np.ndarray:
        """days between primary and secondary dates."""
        days = self.secondary_dates - self.primary_dates
        return days.astype(int)

    @property
    def meta_urls(self)->np.ndarray:
        """metadata urls of LiCSAR. 
        
        metadata includes: 
        
        * E, N, U, hgt 
        * baselines
        * metadata.txt
        * network.png
        * frame_id-poly.txt
        """
        urls = []
        file_names = [
            f"{self.frame_id}.geo.{ENU}.tif" for ENU in ["E", "N", "U", "hgt"]
        ]
        files = file_names + [
            "baselines",
            "metadata.txt",
            "network.png",
            f"{self.frame_id}-poly.txt",
        ]

        for file in files:
            url = f"{self.home_url}/metadata/{file}"
            urls.append(url)

        return np.array(urls)



from __future__ import annotations

import sys
import warnings
import zipfile
from datetime import datetime
from pathlib import Path
from time import sleep

import hyp3_sdk as sdk
import numpy as np
import pandas as pd
from dateutil.parser import parse as parse_date
from tqdm.auto import tqdm

from data_downloader import downloader

from .constants import JOB_TYPE, STATUS_CODE


def ensure_faninsar():
    if "faninsar" in sys.modules:
        return None

    global Pairs
    try:
        from faninsar import Pairs
    except ImportError:
        raise ImportError(
            "FanInSAR package not installed. Please install it using 'pip install FanInSAR'"
        )


def id_of_job(job: sdk.Job) -> str:
    return f"{job.job_id}{job.job_type}"


class Job(sdk.Job):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __hash__(self) -> int:
        return hash(id_of_job(self))

    def __lt__(self, other: "Job") -> bool:
        return id_of_job(self) < id_of_job(other)

    def __eq__(self, other: "Job") -> bool:
        return id_of_job(self) == id_of_job(other)

    def __gt__(self, other: "Job") -> bool:
        return id_of_job(self) > id_of_job(other)

    @staticmethod
    def from_dict(input_dict: dict):
        expiration_time = (
            parse_date(input_dict["expiration_time"])
            if input_dict.get("expiration_time")
            else None
        )
        return Job(
            job_type=input_dict["job_type"],
            job_id=input_dict["job_id"],
            request_time=parse_date(input_dict["request_time"]),
            status_code=input_dict["status_code"],
            user_id=input_dict["user_id"],
            name=input_dict.get("name"),
            job_parameters=input_dict.get("job_parameters"),
            files=input_dict.get("files"),
            logs=input_dict.get("logs"),
            browse_images=input_dict.get("browse_images"),
            thumbnail_images=input_dict.get("thumbnail_images"),
            expiration_time=expiration_time,
            processing_times=input_dict.get("processing_times"),
            credit_cost=input_dict.get("credit_cost"),
        )


class Jobs:
    """A class to manage HyP3 jobs. It provides a pythonic interface to filter
    and select jobs using the numpy and pandas libraries. This class is designed
    to be used as the jobs attribute of the HyP3Service class.
    """

    def __init__(self, jobs: list[sdk.Job]) -> None:
        """Initialize the Jobs class

        Parameters
        ----------
        jobs : list[sdk.Job]
            List of Job objects from HyP3 SDK. You can get the jobs from the
            hyp3_sdk.Batch.jobs attribute.
        """
        self.jobs = pd.Series(
            [Job.from_dict(i.to_dict()) for i in jobs],
            dtype="O",
        )
        (
            self._job_id,
            self._job_type,
            self._request_time,
            self._status_code,
            self._user_id,
            self._name,
            self._job_parameters,
            self._files,
            self._logs,
            self._browse_images,
            self._thumbnail_images,
            self._expiration_time,
            self._processing_times,
            self._credit_cost,
        ) = self._retrieve_jobs()

    def __repr__(self) -> str:
        return f"Jobs({len(self.jobs)})"

    def __len__(self) -> int:
        return len(self.jobs)

    def __add__(self, other: "Jobs") -> "Jobs":
        if not isinstance(other, Jobs):
            raise ValueError("Can only sum Jobs with Jobs instances.")
        return Jobs(pd.concat([self.jobs, other.jobs]))

    def __sub__(self, other: "Jobs") -> "Jobs":
        if not isinstance(other, Jobs):
            raise ValueError("Can only subtract Jobs with Jobs instances.")
        mask = np.isin(self.jobs, other.jobs)
        return Jobs(self.jobs[~mask])

    def __iter__(self):
        return iter(self.jobs)

    def __getitem__(self, key):
        return Jobs(self.jobs[key])

    def _retrieve_jobs(self) -> list[sdk.Job]:
        """Retrieve jobs based on job type and status code

        Parameters
        ----------
        job_type : Literal[job_type]
            Job type to filter by
        status_code : Literal[status_code]
            Status code to filter by
        """
        """Convert the jobs to a pandas DataFrame"""
        job_id = []
        job_type = []
        request_time = []
        status_code = []
        user_id = []
        name = []
        job_parameters = []
        files = []
        logs = []
        browse_images = []
        thumbnail_images = []
        expiration_time = []
        processing_times = []
        credit_cost = []
        for job in self.jobs:
            job_id.append(job.job_id)
            job_type.append(job.job_type)
            request_time.append(job.request_time)
            status_code.append(job.status_code)
            user_id.append(job.user_id)
            name.append(job.name)
            job_parameters.append(job.job_parameters)
            files.append(self._retrieve_files(job.files))
            logs.append(self._retrieve_list(job.logs))
            browse_images.append(self._retrieve_list(job.browse_images))
            thumbnail_images.append(self._retrieve_list(job.thumbnail_images))
            expiration_time.append(job.expiration_time)
            processing_times.append(self._retrieve_list(job.processing_times))
            credit_cost.append(job.credit_cost)

        return (
            job_id,
            job_type,
            request_time,
            status_code,
            user_id,
            name,
            job_parameters,
            files,
            logs,
            browse_images,
            thumbnail_images,
            expiration_time,
            processing_times,
            credit_cost,
        )

    def _retrieve_list(self, val: list | None) -> str | None:
        """Retrieve job from a list"""
        if val is None:
            return np.nan
        if len(val) == 0:
            return np.nan
        return val[0]

    def _ensure_datetime(self, val: datetime | str) -> np.datetime64:
        """Convert a string to a datetime object, otherwise return the object"""
        if isinstance(val, str):
            return pd.to_datetime(val)
        return val

    def _retrieve_files(self, files: list) -> dict:
        """Retrieve files from a list"""
        dict_null = {"filename": np.nan, "s3": np.nan, "size": np.nan, "url": np.nan}
        if files is None:
            return dict_null
        if len(files) == 0:
            return dict_null
        return files[0]

    @property
    def job_type(self) -> np.ndarray:
        """the job type of all jobs"""
        return np.array(self._job_type, dtype=np.str_)

    @property
    def job_id(self) -> np.ndarray:
        """the job ID of all jobs"""
        return np.array(self._job_id, dtype=np.str_)

    @property
    def request_time(self) -> np.ndarray:
        """the request time of all jobs"""
        return np.array(self._request_time, dtype="M8[D]")

    @property
    def status_code(self) -> np.ndarray:
        """the status code of all jobs"""
        return np.array(self._status_code, dtype=np.str_)

    @property
    def user_id(self) -> np.ndarray:
        """the user ID of all jobs"""
        return np.array(self._user_id, dtype=np.str_)

    @property
    def name(self) -> np.ndarray:
        """the name of all jobs"""
        return np.array(self._name, dtype=np.str_)

    @property
    def job_parameters(self) -> np.ndarray:
        """the job parameters of all jobs"""
        return np.array(self._job_parameters, dtype="O")

    @property
    def files(self) -> np.ndarray:
        """the files of all jobs"""
        return np.array(self._files, dtype="O")

    @property
    def file_names(self) -> np.ndarray:
        """the file names of all jobs"""
        return np.array([file["filename"] for file in self._files])

    @property
    def file_urls(self) -> np.ndarray:
        """the file urls of all jobs"""
        return np.array([file["url"] for file in self._files])

    @property
    def file_sizes(self) -> np.ndarray:
        """the file sizes of all jobs"""
        return np.array([file["size"] for file in self._files])

    @property
    def logs(self) -> np.ndarray:
        """the logs of all jobs"""
        return np.array(self._logs, dtype=np.str_)

    @property
    def browse_images(self) -> np.ndarray:
        """the browse images of all jobs"""
        return np.array(self._browse_images)

    @property
    def thumbnail_images(self) -> np.ndarray:
        """the thumbnail images of all jobs"""
        return np.array(self._thumbnail_images, dtype=np.str_)

    @property
    def expiration_time(self) -> np.ndarray:
        """the expiration time of all jobs"""
        return np.array(self._expiration_time, dtype="M8[D]")

    @property
    def processing_times(self) -> np.ndarray:
        """the processing times of all jobs"""
        return np.array(self._processing_times, dtype=np.str_)

    @property
    def credit_cost(self) -> np.ndarray:
        """the credit cost of all jobs"""
        return np.array(self._credit_cost, dtype=np.int_)

    @property
    def total_credit_cost(self) -> int:
        """the total credit cost of all jobs"""
        return np.nansum(self.credit_cost)

    @property
    def frame(self) -> pd.DataFrame:
        """jobs in the form of a pandas DataFrame"""
        df = pd.DataFrame(
            {
                "name": self.name,
                "job_type": self.job_type,
                "status_code": self.status_code,
                "file_names": self.file_names,
                "file_sizes": self.file_sizes,
                "file_urls": self.file_urls,
                "credit_cost": self.credit_cost,
                "request_time": self.request_time,
                "processing_times": self.processing_times,
                "expiration_time": self.expiration_time,
                "job_parameters": self.job_parameters,
                "user_id": self.user_id,
                "logs": self.logs,
                "browse_images": self.browse_images,
                "thumbnail_images": self.thumbnail_images,
            },
            index=self.job_id,
        )
        df.index.name = "job_id"
        return df

    def sel(
        self,
        name: str | None = None,
        job_type: JOB_TYPE | None = None,
        status_code: STATUS_CODE | None = None,
        request_time: datetime | str | slice | None = None,
    ) -> "Jobs":
        """Select jobs based on job type and status code

        Parameters
        ----------
        name : str | None
            Name of the job to filter by
        job_type : JOB_TYPE | None
            Job type to filter by
        status_code : STATUS_CODE | None
            Status code to filter by
        request_time : datetime | str | slice | None
            Request time to filter by. Can be a datetime object, a string, or a slice object. If a slice object is used, the start must be a string or a datetime object, and the stop can be None, a string, or a datetime object. If a string is used, it must be in the format that can be converted to a datetime object using pd.to_datetime. by default None
        """
        if job_type != None and not hasattr(JOB_TYPE, job_type):
            raise ValueError(
                f"Invalid job type: {job_type}. Valid job types are: {JOB_TYPE.variables()}"
            )
        if status_code != None and not hasattr(STATUS_CODE, status_code):
            raise ValueError(
                f"Invalid status code: {status_code}. Valid status codes are: {STATUS_CODE.variables()}"
            )

        mask = np.ones(len(self.frame), dtype=bool)
        if request_time is not None:
            if isinstance(request_time, slice):
                if request_time.start is not None:
                    start = self._ensure_datetime(request_time.start)
                if request_time.stop is None:
                    end = datetime.now().date()
                else:
                    end = self._ensure_datetime(request_time.stop)
                mask = (self.request_time >= start) & (self.request_time <= end)
            if isinstance(request_time, str):
                request_time = self._ensure_datetime(request_time)
                mask = self.request_time == request_time
            if isinstance(request_time, datetime):
                mask = self.request_time == request_time

        if name is not None:
            mask = (self.name == name) & mask
        if job_type is not None:
            mask = (self.job_type == job_type) & mask
        if status_code is not None:
            mask = (self.status_code == status_code) & mask

        return Jobs(self.jobs[mask])

    @property
    def succeeded(self) -> "Jobs":
        """all succeeded jobs (not expired by default)"""
        return self.sel(status_code=STATUS_CODE.SUCCEEDED)

    @property
    def failed(self) -> "Jobs":
        """all failed jobs (not expired by default)"""
        return self.sel(status_code=STATUS_CODE.FAILED)

    @property
    def pending(self) -> "Jobs":
        """all pending jobs (not expired by default)"""
        return self.sel(status_code=STATUS_CODE.PENDING)

    @property
    def running(self) -> "Jobs":
        """all running jobs (not expired by default)"""
        return self.sel(status_code=STATUS_CODE.RUNNING)


class HyP3Service:
    """Class to manage HyP3 user information and jobs"""

    _my_info = None
    _jobs = None

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        prompt: bool = False,
        include_expired=False,
    ):
        """Initialize the HyP3Service class

        Parameters
        ----------
        username, password : str, optional
            Username and password for HyP3
        prompt : bool, optional
            Prompt for the username and password in the terminal, by default False.
        include_expired : bool, optional
            Include expired jobs, by default False
        """
        self.include_expired = include_expired
        self.login(username, password, prompt)

    def __repr__(self) -> str:
        return (
            f"HyP3Service(\n    user_id={self.my_info['user_id']}, "
            f"\n    remaining_credits={self.my_info['remaining_credits']}, "
            f"\n    succeeded={len(self.jobs.succeeded)},"
            f"\n    failed={len(self.jobs.failed)},"
            f"\n    pending={len(self.jobs.pending)},"
            f"\n    running={len(self.jobs.running)}"
            "\n)"
        )

    def __str__(self) -> str:
        return f"Hyp3Service({self.my_info['user_id']}, remaining_credits={self.my_info['remaining_credits']})"

    def flush_jobs(self):
        """Flush the jobs"""
        self._jobs = self._parse_jobs()

    def flush_info(self):
        """Flush the user's information"""
        self._my_info = self.hyp3.my_info()

    def flush(self):
        """Flush the jobs and the user's information"""
        self.flush_jobs()
        self.flush_info()

    def login(
        self,
        username: str | None = None,
        password: str | None = None,
        prompt: bool | None = False,
    ):
        """Login to HyP3. If neither username/password nor prompt is provided,
        it will attempts to use credentials from a ``.netrc`` file. If prompt is True,
        the username and password will be prompted in the terminal. Otherwise, the
        username and password must be provided.

        .. note::

            This method will be called automatically when the class is initialized.
            However, you can change the user by calling this method again with the new
            username and password.

        Parameters
        ----------
        username, password : str, optional
            Username and password for HyP3
        prompt : bool, optional
            Prompt for the username and password in the terminal, by default False.
        """
        self.hyp3 = sdk.HyP3(username=username, password=password, prompt=prompt)
        self.flush()

    @property
    def my_info(self):
        """user's information"""
        return self._my_info

    @property
    def jobs(self) -> Jobs:
        """all jobs (not expired by default, set ``include_expired=True`` to
        include expired jobs)"""
        return self._jobs

    def _parse_jobs(self) -> Jobs:
        """Parse all jobs"""
        batch = self.hyp3.find_jobs().filter_jobs(include_expired=self.include_expired)
        return Jobs(batch.jobs)


class InSARMission:
    """Class to manage ``INSAR_GAMMA`` jobs, which are the interferograms generated
    by GAMMA. This class provides a pythonic interface to submit and download
    ``INSAR_GAMMA`` jobs from HyP3.
    """

    job_type = JOB_TYPE.INSAR_GAMMA

    _pairs_succeed = []
    _pairs_failed = []
    _job_parameters: dict

    def __init__(
        self,
        service: HyP3Service,
        granules: pd.Series,
        job_parameters: dict = {},
    ) -> None:
        """Initialize the InSARJob class

        Parameters
        ----------
        service : HyP3Service
            HyP3Service instance to submit the job and check the submitted jobs.
        granules : pd.Series
            A pandas Series containing the granules information. The index of
            the Series should be the date of the granule, and the values should
            be the granule name.
        job_parameters : dict, optional
            Arguments to be passed to the job, by default {}. You can also change
            the job parameters after initializing the class by job_parameters
            attribute.
        """
        ensure_faninsar()
        self.granules = granules
        self.service = service
        self.job_parameters = job_parameters

        # initialize the batch
        self.batch = sdk.Batch()

    @property
    def jobs_on_service(self) -> Jobs:
        f"""Get the {self.job_type} jobs on the service"""
        self.service.flush()
        jobs_valid = self.service.jobs.sel(job_type=self.job_type)
        return jobs_valid

    @property
    def job_parameters(self) -> dict:
        """Job parameters"""
        return self._job_parameters

    @job_parameters.setter
    def job_parameters(self, job_parameters: dict):
        """Set the job parameters"""
        if not isinstance(job_parameters, dict):
            raise ValueError("job_parameters must be a dictionary.")
        self._job_parameters = job_parameters

    @property
    def pairs_succeed(self) -> Pairs:
        """Pairs that succeeded in the job submission"""
        if len(self._pairs_succeed) == 0:
            return None
        return Pairs(self._pairs_succeed)

    @property
    def pairs_failed(self) -> Pairs:
        """Pairs that failed in the job submission"""
        if len(self._pairs_failed) == 0:
            return None
        return Pairs(self._pairs_failed)

    def jobs_to_pairs(self, jobs: Jobs) -> Pairs:
        f"""Convert {self.job_type} jobs to pairs"""
        pairs = []
        for job in jobs:
            if job.job_type != self.job_type:
                warnings.warn(
                    f"Job type {job.job_type} is not {self.job_type}. Skipping."
                )
                continue
            if len(job.job_parameters["granules"]) != 2:
                warnings.warn(
                    f"Invalid number of granules for job {job.job_id}. Skipping."
                )
                continue
            pair = (
                granule_to_date(job.job_parameters["granules"][0]),
                granule_to_date(job.job_parameters["granules"][1]),
            )
            pairs.append(pair)
        if len(pairs) == 0:
            warnings.warn("No valid pairs found.")
            return None
        return Pairs(pairs)

    def _get_remain_pairs(self, pairs: Pairs, skip_existing: bool = True):
        """Get the remaining pairs to submit"""
        if not skip_existing:
            return pairs

        pairs_exclude = self.jobs_to_pairs(self.jobs_valid)
        if pairs_exclude is None:
            return pairs
        warnings.warn(
            f"Skipping {len(pairs_exclude)} existing pairs already submitted."
        )
        pairs_remain = pairs - pairs_exclude
        return pairs_remain

    def _submit_job(
        self,
        reference,
        secondary,
    ):
        self.batch += self.service.hyp3.submit_insar_job(
            reference, secondary, **self.job_parameters
        )

    def submit_jobs(self, pairs: Pairs, skip_existing: bool = True):
        """Submit the job to HyP3

        Parameters
        ----------
        pairs : Pairs
            Pairs to be submitted to HyP3
        skip_existing : bool, optional
            Whether to skip the existing pairs that have succeeded or are running,
            by default True
        """
        pairs_remain = self._get_remain_pairs(pairs, skip_existing)
        for pair in tqdm(pairs_remain, desc="Submitting jobs"):
            try:
                ref, sec = str(pair).split("_")
                reference, secondary = self.granules[ref], self.granules[sec]

                if reference is None or secondary is None:
                    tqdm.write(f"Granule not found for pair {pair}. Skipping.")
                    self._pairs_failed.append(pair)
                    continue

                def _ensure_granules(granule: pd.Series | str) -> str:
                    """Remove the duplicate pairs"""
                    if isinstance(granule, pd.Series):
                        tqdm.write(
                            f"Multiple granules found for pair {pair}: {[i for i in granule]}."
                            "First one will be used."
                        )
                        granule = granule[0]
                    return granule

                reference = _ensure_granules(reference)
                secondary = _ensure_granules(secondary)
                self._submit_job(reference, secondary)
                self._pairs_succeed.append(pair)
            except Exception as e:
                params = {"granule1": reference, "granule2": secondary}
                params.update(self.job_parameters)

                tqdm.write(
                    f"Failed to submit job for pair {pair}. {e}."
                    f"\nJob parameters: {params}"
                )
                self._pairs_failed.append(pair)

    def _scan_interferograms(self, home_dir: Path | str) -> list[str]:
        """Scan the local directory for the interferograms"""
        home_dir = Path(home_dir)
        return [i.stem for i in home_dir.glob("*") if i.is_dir()]

    def _download_jobs(
        self,
        output_dir: Path | str,
        name: str | None = None,
        request_time: datetime | str | slice | None = None,
        unzip: bool = True,
        remove_zip: bool = True,
        overwrite: bool = False,
    ):
        f"""Download the {self.job_type} jobs from HyP3"""
        jobs = self.jobs_on_service.sel(name=name, request_time=request_time).succeeded
        for file_name, url in tqdm(
            zip(jobs.file_names, jobs.file_urls),
            desc="Downloading jobs",
            total=len(jobs),
        ):
            local_ifgs = self._scan_interferograms(output_dir)
            if Path(file_name).stem in local_ifgs and not overwrite:
                tqdm.write(f"Interferogram {file_name} already exists. Skipping.")
                continue
            try:
                downloader.download_data(url, output_dir, file_name)
                if unzip:
                    unzip_file(output_dir, file_name, remove_zip, overwrite)
            except Exception as e:
                tqdm.write(f"Failed to download file {file_name}. {e}")

    def download_jobs(
        self,
        output_dir: Path | str,
        name: str | None = None,
        request_time: datetime | str | slice | None = None,
        unzip: bool = True,
        remove_zip: bool = True,
        overwrite: bool = False,
        wait_running=True,
        wait_minutes=60,
        retry=3,
    ):
        f"""Download the {self.job_type} jobs from HyP3

        Parameters
        ----------
        output_dir : Path | str
            Output directory to save the files
        name : str, optional
            Name of submitted jobs to filter by, by default None
        request_time : datetime | str | slice, optional
            Request time of submitted jobs to filter by. Can be a datetime object,
            a string, or a slice object. If a slice object is used, the start
            must be a string or a datetime object, and the stop can be None, a
            string. If a string is used, it must be in the format that can be
            converted to a datetime
        unzip : bool, optional
            Whether to unzip the files, by default True
        remove_zip : bool, optional
            Whether to remove the zip files after unzipping, by default True
        overwrite : bool, optional
            Whether to overwrite the existing files when unzipping. If False, The
            interferogram folders that are already unzipped will not be downloaded
            again, by default False.
        wait_running : bool, optional
            Whether to wait for the jobs that are still running, by default True
        wait_minutes : int, optional
            Time to wait for the jobs to finish, by default 60 (1 hour)
        retry : int, optional
            Number of times to retry the download, by default 3
        """
        count = 0
        while True:
            self._download_jobs(
                output_dir, name, request_time, unzip, remove_zip, overwrite
            )
            count += 1
            if count >= retry:
                break
            # check if there are still running jobs
            jobs = self.jobs_on_service.sel(name=name, request_time=request_time)
            if len(jobs.running) == 0:
                break
            if not wait_running:
                warnings.warn(
                    "Some jobs are still running. You may need to download them later."
                )
                break
            wait_minutes = int(wait_minutes)
            tqdm.write(
                "Downloading jobs finished. But some jobs are still running."
                f"\nA new download will be attempted in {wait_minutes} minutes."
                "\nYou can stop the process by pressing Ctrl+C."
            )
            sleep(wait_minutes * 60)


class InSARBurstMission(InSARMission):
    """Class to manage ``INSAR_ISCE_BURST`` jobs, which are the burst interferograms
    generated by ISCE. This class provides a pythonic interface to submit and
    download ``INSAR_ISCE_BURST`` jobs from HyP3.
    """

    job_type = JOB_TYPE.INSAR_ISCE_BURST

    def __init__(
        self, service: HyP3Service, granules: pd.Series, job_parameters: dict = {}
    ) -> None:
        super().__init__(service, granules, job_parameters)

    def _submit_job(
        self,
        reference,
        secondary,
    ):
        self.batch += self.service.hyp3.submit_insar_isce_burst_job(
            reference, secondary, **self.job_parameters
        )


def granule_to_date(granule: str):
    """Convert granule to date"""
    return pd.to_datetime(granule.split("_")[5])


def unzip_file(output_dir, file_name, remove_zip=True, overwrite=False):
    try:
        zip_file = Path(output_dir) / file_name
        unzip_dir = Path(output_dir) / Path(file_name).stem
        if not overwrite and unzip_dir.exists():
            warnings.warn(f"Directory {unzip_dir} already exists. Skipping.")
            return None
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            zip_ref.extractall(output_dir)
        if remove_zip:
            zip_file.unlink()
    except Exception as e:
        warnings.warn(f"Error in unzipping {zip_file}: {e}")



class BaseConstants:
    def __init__(self) -> None:
        pass

    @classmethod
    def variables(cls):
        """Returns a list of all available variables"""
        variables = [
            prop
            for prop in dir(cls)
            if not prop.startswith("_") and not callable(getattr(cls, prop))
        ]

        return variables

class JOB_TYPE(BaseConstants):
    AUTORIFT = "AUTORIFT"
    RTC_GAMMA = "RTC_GAMMA"
    INSAR_GAMMA = "INSAR_GAMMA"
    INSAR_ISCE_BURST = "INSAR_ISCE_BURST"


class STATUS_CODE(BaseConstants):
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    RUNNING = "RUNNING"
    PENDING = "PENDING"



import setuptools

with open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="data_downloader",
    version="1.1",
    author="fanchegyan",
    author_email="fanchy14@lzu.edu.cn",
    description="Make downloading scientific data much easier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Fanchengyan/data_downloader",
    packages=setuptools.find_packages(),
    install_requires=[
        "httpx >= 0.4.0",
        "requests",
        "tqdm",
        "setuptools",
        "beautifulsoup4",
        "nest_asyncio",
        "python-dateutil",
        "browser-cookie3",
        "hyp3_sdk",
        "pandas",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



from __future__ import annotations
import asyncio
import datetime as dt
import multiprocessing as mp
import os
import selectors
import time
from netrc import netrc
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

import browser_cookie3 as bc
import httpx
import nest_asyncio
import requests
from dateutil.parser import parse
from tqdm.auto import tqdm

nest_asyncio.apply()


def get_url_host(url):
    """Returns the url host for a given url"""
    ri = urlparse(url)
    # Strip port numbers from netloc. This weird `if...encode`` dance is
    # used for Python 3.2, which doesn't support unicode literals.
    splitstr = b":"
    if isinstance(url, str):
        splitstr = splitstr.decode("ascii")
    host = ri.netloc.split(splitstr)[0]
    return host


def get_netrc_auth(url):
    """Returns the Requests tuple auth for a given url from .netrc"""
    host = get_url_host(url)
    _netrc = Netrc().authenticators(host)

    if _netrc:
        # Return with login / password
        login_i = 0 if _netrc[0] else 1
        return (_netrc[login_i], _netrc[2])


class Netrc(netrc):
    """a class managing records in .netrc file"""

    def __init__(self, file: Optional[Union[str, Path]] = None):
        if file is None:
            file = Path("~/.netrc").expanduser()
        else:
            file = Path(file)
        self.file = file
        if not file.exists():
            open(self.file, "w").close()

        netrc.__init__(self, file)

    def _info_to_file(self):
        rep = self.__repr__()
        with open(self.file, "w") as f:
            f.write(rep)

    def _update_info(self):
        with open(self.file) as fp:
            self._parse(self.file, fp, False)

    def add(self, host, login, password, account=None, overwrite=False):
        """add a record

        Will do nothing if host exists in .netrc file unless set overwrite=True
        """
        if host in self.hosts and not overwrite:
            tqdm.write(
                f">>> Warning: {host} existed, nothing will be done."
                + " If you want to overwrite the existed record, set overwrite=True"
            )
        else:
            self.hosts.update({host: (login, account, password)})
            self._info_to_file()
            self._update_info()

    def remove(self, host):
        """remove a record by host"""
        self.hosts.pop(host)
        self._info_to_file()
        self._update_info()

    def clear(self):
        """remove all records"""
        self.hosts = {}
        self._info_to_file()
        self._update_info()


def _parse_file_name(response):
    """parse the file_name from the headers of web response or url"""

    if "Content-disposition" in response.headers:
        file_name = (
            response.headers["Content-disposition"]
            .split("filename=")[1]
            .strip('"')
            .strip("'")
        )
    else:
        file_name = os.path.basename(urlparse(str(response.url)).path)
    return file_name


def _unit_formater(size, suffix):
    prefixs = ["", "k", "M", "G", "T"]
    idx = 0
    while size / 1024 >= 1:
        size = size / 1024
        idx += 1
        if idx == 4:
            break

    return f"{size:.2f}{prefixs[idx]}{suffix}"


def _new_file_from_web(r, file_path):
    """whether have new file from the website"""
    try:
        time_remote = parse(r.headers.get("Last-Modified"))
        time_local = dt.datetime.fromtimestamp(
            os.path.getmtime(file_path), dt.timezone.utc
        )
        return time_remote > time_local
    except:
        return False


def _get_cookiejar(authorize_from_browser):
    cj = None
    if authorize_from_browser:
        try:
            cj = bc.load()
        except:
            raise EnvironmentError(
                "Could not load cookie from browser. "
                "Please login in website via browser before run this code"
                "\n  So far the following browsers are supported: "
                "Chrome,Firefox, Opera, Edge, Chromium"
            )
    return cj


def _handle_status(r, url, local_size, file_name, file_path):
    # returns (True, '') : downloaded entirely
    # returns (False,'') : error! break download
    # returns (False, url) : 301,302
    # returns None: continue to download

    global support_resume, pbar, remote_size

    if r.status_code in [206, 416]:
        support_resume = True
        remote_size = int(r.headers["Content-Range"].rsplit("/")[-1])

        # init process bar
        if _new_file_from_web(r, file_path):
            tqdm.write(
                f"There is a new file from {url}"
                f"{file_name} is ready to be downloaded again"
            )
            os.remove(file_path)
        elif local_size < remote_size:
            pbar = tqdm(
                initial=local_size,
                total=remote_size,
                unit="B",
                unit_scale=True,
                dynamic_ncols=True,
                desc=Path(file_name).name,
            )
        else:
            tqdm.write(f"{file_name} was downloaded entirely. skiping download")
            return True, ""
    elif r.status_code == 200:
        # know the total size, then delete the file that wasn't downloaded entirely and redownload it.
        if "Content-length" in r.headers:
            remote_size = int(r.headers["Content-length"])

            if _new_file_from_web(r, file_path):
                tqdm.write(
                    f"There is a new file from {url}"
                    f"{file_name} is ready to be downloaded again"
                )
                os.remove(file_path)
            elif 0 < local_size < remote_size:
                tqdm.write(f"  Detect {file_name} wasn't downloaded entirely")
                tqdm.write(
                    "  The website not supports resuming breakpoint."
                    " Prepare to remove the local file and redownload..."
                )
                os.remove(file_path)
            elif local_size > remote_size:
                tqdm.write(
                    "Detected the local file is larger than the server file. "
                    " Prepare to remove local the file and redownload..."
                )
                os.remove(file_path)
            elif local_size == remote_size:
                tqdm.write(f"{file_name} was downloaded entirely. skiping download")
                return True, ""
        # don't know the total size, warning user if detect the file was downloaded.
        else:
            if os.path.exists(file_path):
                tqdm.write(
                    f">>> Warning: Detect the {file_name} was downloaded,"
                    " but can't parse the it's size from website\n"
                    f"    If you know it wasn't downloaded entirely, delete "
                    "it and redownload it again. skiping download..."
                )
                return True, ""
    elif r.status_code == 202:
        tqdm.write(
            ">>> The server has accepted your request but has not yet processed it. "
            "Please redownload it later"
        )
        return False, ""
    elif r.status_code in [301, 302]:
        url_new = r.headers["Location"]
        tqdm.write(f">>> Waring: the website has redirected to {url_new}")
        return False, url_new
    elif r.status_code == 401:
        tqdm.write(
            ">>> Authorization failed! Please check your username and password in Netrc "
            "Or authorizing by browser and set the parameter `authorize_from_browser` to `True`"
        )
        return False, ""
    elif r.status_code == 403:
        tqdm.write(
            ">>> Forbidden! Access to the requested resource was denied by the server"
        )
        return False, ""
    else:
        tqdm.write(
            f'  Download file from "{url}" failed,'
            f" The service returns the HTTP Status Code: {r.status_code}"
        )
        return False, ""


def _download_data_httpx(
    url,
    folder=None,
    file_name=None,
    client=None,
    follow_redirects=True,
    retry=0,
    authorize_from_browser=False,
):
    """Download a single file.

    Parameters:
    -----------
    url: str
        url of web file
    folder: str
        the folder to store output files. Default current folder.
    authorize_from_browser: bool
        Whether to load cookies used by your web browser for authorization.
        This means you can use python to download data by logging in to website
        via browser (So far the following browsers are supported: Chrome,Firefox,
        Opera, Edge, Chromium"). It will be very useful when website doesn't support
        "HTTP Basic Auth". Default is False.
    file_name: str
        the file name. If None, will parse from web response or url.
        file_name can be the absolute path if folder is None.
    client: httpx.Client() object
        client maintaining connection. Default None
    follow_redirects: bool
        Enables or disables HTTP redirects
    retry: int
        number of reconnection when status code is 503
    """
    # init parameters
    global support_resume, pbar, remote_size

    support_resume = False
    headers = {"Range": "bytes=0-4"}
    if not client:
        client = httpx

    cj = _get_cookiejar(authorize_from_browser)

    r = client.get(
        url, headers=headers, timeout=120, follow_redirects=follow_redirects, cookies=cj
    )
    r.close()

    if file_name is None:
        file_name = _parse_file_name(r)

    if folder is not None:
        file_path = os.path.join(folder, file_name)
    else:
        file_path = os.path.abspath(file_name)

    local_size = os.path.getsize(file_path) if os.path.exists(file_path) else 0

    result = _handle_status(r, url, local_size, file_name, file_path)
    if result:
        status, url_new = result
        if status:  # downloaded entirely
            return True
        elif status == False:
            if url_new:  # 301,302
                return _download_data_httpx(
                    url_new,
                    folder=folder,
                    file_name=file_name,
                    authorize_from_browser=authorize_from_browser,
                    follow_redirects=True,
                    client=client,
                )
            elif retry > 0:
                return _download_data_httpx(
                    url,
                    folder=folder,
                    file_name=file_name,
                    authorize_from_browser=authorize_from_browser,
                    client=client,
                    retry=retry - 1,
                )
            else:  # error! break download
                return False

    # begin downloading
    if support_resume:
        headers["Range"] = f"bytes={local_size}-{remote_size}"
    else:
        headers = None

    with client.stream("GET", url, headers=headers, timeout=120, cookies=cj) as r:
        with open(file_path, "ab") as f:
            time_start_realtime = time_start = time.time()
            for chunk in r.iter_raw():
                if chunk:
                    size_add = len(chunk)
                    local_size += size_add
                    f.write(chunk)
                    f.flush()
                if support_resume:
                    pbar.update(size_add)
                else:
                    time_end_realtime = time.time()
                    time_span = time_end_realtime - time_start_realtime
                    if time_span > 1:
                        speed_realtime = size_add / time_span
                        tqdm.write(
                            "  Downloading {} [Speed: {} | Size: {}]".format(
                                file_name,
                                _unit_formater(speed_realtime, "B/s"),
                                _unit_formater(local_size, "B"),
                            ),
                            end="\r",
                        )
                        time_start_realtime = time_end_realtime
            if not support_resume:
                time_cost = time.time() - time_start
                speed = local_size / time_cost if time_cost > 0 else 0
                tqdm.write(
                    "  Finish downloading {} [Speed: {} | Total Size: {}]".format(
                        file_name,
                        _unit_formater(speed, "B/s"),
                        _unit_formater(local_size, "B"),
                    )
                )
    return True


def _download_data_requests(
    url,
    folder=None,
    file_name=None,
    client=None,
    follow_redirects=True,
    retry=0,
    authorize_from_browser=False,
):
    """Download a single file.

    Parameters:
    -----------
    url: str
        url of web file
    folder: str
        the folder to store output files. Default current folder.
    authorize_from_browser: bool
        Whether to load cookies used by your web browser for authorization.
        This means you can use python to download data by logging in to website
        via browser (So far the following browsers are supported: Chrome,Firefox,
        Opera, Edge, Chromium"). It will be very useful when website doesn't support
        "HTTP Basic Auth". Default is False.
    file_name: str
        the file name. If None, will parse from web response or url.
        file_name can be the absolute path if folder is None.
    client: requests.Session() object
        client maintaining connection. Default None
    follow_redirects: bool
        Enables or disables HTTP redirects
    retry: int
        number of reconnection when status code is 503
    """
    # init parameters
    global support_resume, pbar, remote_size

    support_resume = False
    headers = {"Range": "bytes=0-4"}
    if not client:
        client = requests

    cj = _get_cookiejar(authorize_from_browser)

    r = client.get(
        url, headers=headers, timeout=120, allow_redirects=follow_redirects, cookies=cj
    )
    r.close()

    if file_name is None:
        file_name = _parse_file_name(r)

    if folder is not None:
        file_path = os.path.join(folder, file_name)
    else:
        file_path = os.path.abspath(file_name)

    local_size = os.path.getsize(file_path) if os.path.exists(file_path) else 0

    result = _handle_status(r, url, local_size, file_name, file_path)
    if result:
        status, url_new = result
        if status:  # downloaded entirely
            return True
        elif status == False:
            if url_new:  # 301,302
                return _download_data_requests(
                    url_new,
                    folder=folder,
                    file_name=file_name,
                    authorize_from_browser=authorize_from_browser,
                    follow_redirects=True,
                    client=client,
                )
            elif retry > 0:
                return _download_data_requests(
                    url,
                    folder=folder,
                    file_name=file_name,
                    authorize_from_browser=authorize_from_browser,
                    client=client,
                    retry=retry - 1,
                )
            else:  # error! break download
                return False

    # begin downloading
    if support_resume:
        headers["Range"] = f"bytes={local_size}-{remote_size}"
    else:
        headers = None

    r = client.get(url, headers=headers, stream=True, timeout=120, cookies=cj)
    with open(file_path, "ab") as f:
        time_start_realtime = time_start = time.time()
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                size_add = len(chunk)
                local_size += size_add
                f.write(chunk)
                f.flush()
            if support_resume:
                pbar.update(size_add)
            else:
                time_end_realtime = time.time()
                time_span = time_end_realtime - time_start_realtime
                if time_span > 1:
                    speed_realtime = size_add / time_span
                    tqdm.write(
                        "  Downloading {} [Speed: {} | Size: {}]".format(
                            file_name,
                            _unit_formater(speed_realtime, "B/s"),
                            _unit_formater(local_size, "B"),
                        ),
                        end="\r",
                    )
                    time_start_realtime = time_end_realtime
        if not support_resume:
            time_cost = time.time() - time_start
            speed = local_size / time_cost if time_cost > 0 else 0
            tqdm.write(
                "  Finish downloading {} [Speed: {} | Total Size: {}]".format(
                    file_name,
                    _unit_formater(speed, "B/s"),
                    _unit_formater(local_size, "B"),
                )
            )

    r.close()
    return True


def download_data(
    url,
    folder=None,
    file_name=None,
    client=None,
    engine="requests",
    follow_redirects=True,
    retry=0,
    authorize_from_browser=False,
):
    """Download a single file.

    Parameters:
    -----------
    url: str
        url of web file
    folder: str
        the folder to store output files. Default current folder.
    file_name: str
        the file name. If None, will parse from web response or url.
        file_name can be the absolute path if folder is None.
    client: requests.Session() for `requests` engine or httpx.Client() for `httpx` engine
        client maintaining connection. Default None
    engine: one of ["requests","httpx"]
        engine for downloading
    follow_redirects: bool
        Enables or disables HTTP redirects
    retry: int
        number of reconnection when status code is 503
    authorize_from_browser: bool
        Whether to load cookies used by your web browser for authorization.
        This means you can use python to download data by logging in to website
        via browser (So far the following browsers are supported: Chrome,Firefox,
        Opera, Edge, Chromium"). It will be very useful when website doesn't support
        "HTTP Basic Auth". Default is False.
    """

    if engine == "requests":
        _download_data_requests(
            url,
            folder,
            file_name,
            client,
            follow_redirects,
            retry,
            authorize_from_browser,
        )
    elif engine == "httpx":
        _download_data_httpx(
            url,
            folder,
            file_name,
            client,
            follow_redirects,
            retry,
            authorize_from_browser,
        )
    else:
        raise ValueError('engine must be one of ["requests","httpx"]')


def download_datas(
    urls,
    folder=None,
    file_names=None,
    engine="requests",
    authorize_from_browser=False,
    desc="",
):
    """download data from a list like object which containing urls.
    This function will download files one by one.

    Parameters:
    -----------
    urls:  iterator
        iterator contains urls
    folder: str
        the folder to store output files. Default current folder.
    engine: one of ["requests","httpx"]
        engine for downloading
    file_names: iterator
        iterator contains names of files. Leaving it None if you want the program to parse
        them from website. file_names can contain the absolute paths if folder is None.
    authorize_from_browser: bool
        Whether to load cookies used by your web browser for authorization.
        This means you can use python to download data by logging in to website
        via browser (So far the following browsers are supported: Chrome,Firefox,
        Opera, Edge, Chromium"). It will be very useful when website doesn't support
        "HTTP Basic Auth". Default is False.
    desc: str
        description of data downloading

    Examples:
    ---------

    >>> from data_downloader import downloader
    
    specify the urls and folder
    
    >>> urls=['http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20141211/20141117_20141211.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150221/20141024_20150221.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150128/20141024_20150128.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150128/20141024_20150128.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141211_20150128/20141211_20150128.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20150317/20141117_20150317.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20150221/20141117_20150221.geo.cc.tif']
    >>> folder = 'D:\\data'
    
    download data from urls and store them in folder
    
    >>> downloader.download_datas(urls,folder)
    """
    if engine == "requests":
        client = requests.Session()
    elif engine == "httpx":
        client = httpx.Client(timeout=None)
    else:
        raise ValueError('engine must be one of ["requests","httpx"]')

    desc = ">>> Total | " + desc.title()
    for i, url in enumerate(tqdm(urls, unit="files", dynamic_ncols=True, desc=desc)):
        if file_names is not None:
            download_data(
                url,
                folder,
                file_name=file_names[i],
                client=client,
                engine=engine,
                authorize_from_browser=authorize_from_browser,
            )
        else:
            download_data(
                url,
                folder,
                client=client,
                engine=engine,
                authorize_from_browser=authorize_from_browser,
            )


def _mp_download_data(args):
    return download_data(*args)


def mp_download_datas(
    urls,
    folder=None,
    file_names=None,
    ncore=None,
    desc="",
    follow_redirects=True,
    retry=0,
    engine="requests",
    authorize_from_browser=False,
):
    """download data from a list like object which containing urls.
    This function will download multiple files simultaneously using multiprocess.

    Parameters:
    -----------
    urls:  iterator
        iterator contains urls
    folder: str
        the folder to store output files. Default current folder.
    engine: one of ["requests","httpx"]
        engine for downloading
    file_names: iterator
        iterator contains names of files. Leaving it None if you want the program to parse
        them from website. file_names can contain the absolute paths if folder is None.
    ncore: int
        Number of cores for parallel processing. If ncore is None then the number returned
        by os.cpu_count() is used. Default None.
    desc: str
        description of data downloading
    authorize_from_browser: bool
        Whether to load cookies used by your web browser for authorization.
        This means you can use python to download data by logging in to website
        via browser (So far the following browsers are supported: Chrome,Firefox,
        Opera, Edge, Chromium"). It will be very useful when website doesn't support
        "HTTP Basic Auth". Default is False.

    Examples:
    ---------

    >>> from data_downloader import downloader

    specify the urls and folder
    
    >>> urls=['http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20141211/20141117_20141211.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150221/20141024_20150221.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150128/20141024_20150128.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150128/20141024_20150128.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141211_20150128/20141211_20150128.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20150317/20141117_20150317.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20150221/20141117_20150221.geo.cc.tif']
    >>> folder = 'D:\\data'
    
    download data from urls and store them in folder
    
    >>> downloader.mp_download_datas(urls,folder)
    """
    if ncore == None:
        ncore = os.cpu_count()
    else:
        ncore = int(ncore)
    tqdm.write(f">>> {ncore} parallel downloading")

    desc = ">>> Total | " + desc.title()
    pbar = tqdm(total=len(urls), desc=desc, dynamic_ncols=True)

    with mp.Pool(ncore) as pool:
        if file_names is not None:
            args = [
                (
                    urls[i],
                    folder,
                    file_names[i],
                    None,
                    engine,
                    follow_redirects,
                    retry,
                    authorize_from_browser,
                )
                for i in range(len(urls))
            ]  # Need to put other parameters in right places
        else:
            args = [
                (
                    urls[i],
                    folder,
                    file_names,
                    None,
                    engine,
                    follow_redirects,
                    retry,
                    authorize_from_browser,
                )
                for i in range(len(urls))
            ]

        for i in pool.imap_unordered(_mp_download_data, args):
            pbar.update()


async def _download_data(
    client,
    url,
    folder=None,
    file_name=None,
    follow_redirects=True,
    retry=0,
    authorize_from_browser=False,
):
    global support_resume, pbar, remote_size

    headers = {"Range": "bytes=0-4"}
    support_resume = False

    cj = _get_cookiejar(authorize_from_browser)
    # auth = get_netrc_auth(url)

    r = await client.get(
        url, headers=headers, timeout=120, follow_redirects=follow_redirects, cookies=cj
    )
    # r.close()
    # r = await client.head(url, headers=headers, auth=auth, timeout=120)
    if file_name is None:
        file_name = _parse_file_name(r)

    if folder is not None:
        if not os.path.exists(folder):
            os.makedirs(folder)
        file_path = os.path.join(folder, file_name)
    else:
        file_path = os.path.abspath(file_name)

    local_size = os.path.getsize(file_path) if os.path.exists(file_path) else 0

    result = _handle_status(r, url, local_size, file_name, file_path)
    if result:
        status, url_new = result
        if status:  # downloaded entirely
            return True
        elif status == False:
            if url_new:  # 301,302
                return await _download_data(
                    client,
                    url_new,
                    folder=folder,
                    authorize_from_browser=authorize_from_browser,
                    file_name=file_name,
                    follow_redirects=True,
                )
            elif retry > 0:
                return await _download_data(
                    client,
                    url,
                    folder=folder,
                    authorize_from_browser=authorize_from_browser,
                    file_name=file_name,
                    retry=retry - 1,
                )
            else:  # error! break download
                return False

    # begin download
    if support_resume:
        headers["Range"] = f"bytes={local_size}-{remote_size}"
    else:
        headers = None
    auth = get_netrc_auth(get_url_host(url))
    async with client.stream(
        "GET", url, headers=headers, auth=auth, timeout=None, cookies=cj
    ) as r:
        with open(file_path, "ab") as f:
            time_start_realtime = time_start = time.time()

            async for chunk in r.aiter_bytes():
                size_add = len(chunk)
                local_size += size_add
                f.write(chunk)
                f.flush()
                if support_resume:
                    pbar.update(size_add)
                else:
                    time_end_realtime = time.time()
                    time_span = time_end_realtime - time_start_realtime
                    if time_span > 1:
                        speed_realtime = size_add / time_span
                        tqdm.write(
                            "Downloading {} [Speed: {} | Size: {}]".format(
                                file_name,
                                _unit_formater(speed_realtime, "B/s"),
                                _unit_formater(local_size, "B"),
                            ),
                            end="\r",
                        )
                        time_start_realtime = time_end_realtime
            if not support_resume:
                speed = local_size / (time.time() - time_start)
                tqdm.write(
                    "Finish downloading {} [Speed: {} | Total Size: {}]".format(
                        file_name,
                        _unit_formater(speed, "B/s"),
                        _unit_formater(local_size, "B"),
                    )
                )
            # r.close()
            return True


async def creat_tasks(
    urls,
    folder,
    authorize_from_browser,
    file_names,
    limit,
    desc,
    follow_redirects,
    retry,
):
    limits = httpx.Limits(max_keepalive_connections=limit, max_connections=limit)
    async with httpx.AsyncClient(limits=limits, timeout=None, verify=False) as client:
        if file_names is not None:
            tasks = [
                asyncio.ensure_future(
                    _download_data(
                        client,
                        url,
                        folder,
                        authorize_from_browser=authorize_from_browser,
                        file_name=file_names[i],
                        follow_redirects=follow_redirects,
                        retry=retry,
                    )
                )
                for i, url in enumerate(urls)
            ]
        else:
            tasks = [
                asyncio.ensure_future(
                    _download_data(
                        client,
                        url,
                        folder,
                        authorize_from_browser=authorize_from_browser,
                        follow_redirects=follow_redirects,
                        retry=retry,
                    )
                )
                for url in urls
            ]

        # Total process bar
        tasks_iter = asyncio.as_completed(tasks)
        desc = ">>> Total | " + desc.title()
        pbar = tqdm(tasks_iter, total=len(urls), desc=desc, dynamic_ncols=True)
        for coroutine in pbar:
            await coroutine


def async_download_datas(
    urls,
    folder=None,
    file_names=None,
    limit=30,
    desc="",
    follow_redirects=True,
    retry=0,
    authorize_from_browser=False,
):
    """Download multiple files simultaneously.

    Parameters:
    -----------
    urls:  iterator
        iterator contains urls
    folder: str
        the folder to store output files. Default current folder.
    authorize_from_browser: bool
        Whether to load cookies used by your web browser for authorization.
        This means you can use python to download data by logging in to website
        via browser (So far the following browsers are supported: Chrome,Firefox,
        Opera, Edge, Chromium"). It will be very useful when website doesn't support
        "HTTP Basic Auth". Default is False.
    file_names: iterator
        iterator contains names of files. Leaving it None if you want the program
        to parse them from website. file_names can contain the absolute paths if folder is None.
    limit: int
        the number of files downloading simultaneously
    desc: str
        description of datas downloading
    follow_redirects: bool
        Enables or disables HTTP redirects
    retry: int
        number of reconnection when status code is 503

    Example:
    ---------

    >>> from data_downloader import downloader
    
    specify the urls and folder
    
    >>> urls=['http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20141211/20141117_20141211.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150221/20141024_20150221.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150128/20141024_20150128.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141024_20150128/20141024_20150128.geo.unw.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141211_20150128/20141211_20150128.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20150317/20141117_20150317.geo.cc.tif',
    'http://gws-access.ceda.ac.uk/public/nceo_geohazards/LiCSAR_products/106/106D_05049_131313/interferograms/20141117_20150221/20141117_20150221.geo.cc.tif']
    >>> folder = 'D:\\data'
    
    download data from urls and store them in folder
    
    >>> downloader.async_download_datas(urls,folder,None,desc='interferograms')
    """
    # solve the loop close  Error for python 3.8.x in windows platform
    selector = selectors.SelectSelector()
    loop = asyncio.SelectorEventLoop(selector)
    try:
        loop.run_until_complete(
            creat_tasks(
                urls,
                folder,
                authorize_from_browser,
                file_names,
                limit,
                desc,
                follow_redirects,
                retry,
            )
        )
    finally:
        loop.close()


async def _is_response_staus_ok(client, url, authorize_from_browser, timeout):
    cj = _get_cookiejar(authorize_from_browser)
    try:
        r = await client.head(url, timeout=timeout, cookies=cj)
        r.close()
        if r.status_code == httpx.codes.OK:
            return True
        else:
            return False
    except:
        return False


async def creat_tasks_status_ok(urls, limit, authorize_from_browser, timeout):
    limits = httpx.Limits(max_keepalive_connections=limit, max_connections=limit)
    async with httpx.AsyncClient(limits=limits, timeout=None) as client:
        tasks = [
            asyncio.create_task(
                _is_response_staus_ok(client, url, authorize_from_browser, timeout)
            )
            for url in urls
        ]
        status_ok = await asyncio.gather(*tasks)

    return status_ok


def status_ok(urls, limit=200, authorize_from_browser=False, timeout=60):
    """Simultaneously detecting whether the given links are accessible.

    Parameters
    ----------
    urls: iterator
        iterator contains urls
    limit: int
        the number of urls connecting simultaneously
    authorize_from_browser: bool
        Whether to load cookies used by your web browser for authorization.
        This means you can use python to download data by logging in to website
        via browser (So far the following browsers are supported: Chrome,Firefox,
        Opera, Edge, Chromium"). It will be very useful when website doesn't support
        "HTTP Basic Auth". Default is False.
    timeout: int
        Request to stop waiting for a response after a given number of seconds

    Return:
    ------
    a list of results (True or False)

    Example:
    -------
    ```python
    from data_downloader import downloader
    import numpy as np

    urls = np.array(['https://www.baidu.com',
    'https://www.bai.com/wrongurl',
    'https://cn.bing.com/',
    'https://bing.com/wrongurl',
    'https://bing.com/'] )

    status_ok = downloader.status_ok(urls)
    urls_accessible = urls[status_ok]
    tqdm.write(urls_accessible)
    ```
    """
    # solve the loop close  Error for python 3.8.x in windows platform
    selector = selectors.SelectSelector()
    loop = asyncio.SelectorEventLoop(selector)
    try:
        status_ok = loop.run_until_complete(
            creat_tasks_status_ok(urls, limit, authorize_from_browser, timeout)
        )
    # Zero-sleep to allow underlying connections to close
    finally:
        loop.close()

    return status_ok
